// ------------------------------------------------------------------------------------------------------
// LightningChart® example code: Demo shows a record volume of data, over 1 Billion points in real-time. 
//
// If you need any assistance, or notice error in this example code, please contact support@lightningchart.com. 
//
// Permission to use this code in your application comes with LightningChart® license. 
//
// https://www.arction.com | support@lightningchart.com | sales@lightningchart.com
//
// © Arction Ltd 2009-2022. All rights reserved.  
// ------------------------------------------------------------------------------------------------------
using Arction.Wpf.Charting.Axes;
using Arction.Wpf.Charting;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Arction.Wpf.Charting.SeriesXY;
using System.Windows.Threading;
using Arction.Wpf.Charting.Views.View3D;
using Arction.Wpf.SignalProcessing;
using Arction.Wpf.Charting.Series3D;
using Shape = Arction.Wpf.Charting.Shape;
using System.Diagnostics;
using System.Reflection;
using System.IO;
using Arction.Wpf.Charting.Maps;
using Arction.Wpf.Charting.Views.ViewXY;
using Arction.Wpf.Charting.EventMarkers;
using System.Threading;
using Arction.Wpf.Charting.ChartManager;
using Arction.Wpf.Charting.Annotations;

namespace InteractiveExamples
{
    public partial class ExampleEEGDataVisualizationShowcase : Window, IDisposable
    {

        bool _running = true;
        const double _XMin = -65;
        const double _XMax = 65;
        const double _YMin = -100;
        const double _YMax = 100;
        const int _InterpolatedResolutionX = 80;
        const int _InterpolatedResolutionY = 110;
        int _iDataBufferSize = 35;
        int _iWinSizeForFFT = 500;
        private double _samplingFrequency = 1000;
        private int _iCutOffFR = 100;
        Reader _reader = new Reader($@"{Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData)}\Arction\Content\Output3mins.csv");
        SpectrumCalculator _spectrumCalculator;
        string _Head = $@"{Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData)}\Arction\Content\free_head.obj";
        string _Brains = $@"{Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData)}\Arction\Content\brain.obj";
        List<LightningChart> _ChartList = new List<LightningChart>();
        int _EEGchannelsCount = 21;
        double[][] _buffer;
        double[] _buffer04;
        double[] _buffer48;
        double[] _buffer812;
        double[] _buffer1232;
        double[] _buffer32100;
        double[] _selected;
        int _round = 0;
        GeometryAdditionalData[][] _gads2D;
        SeriesPoint[] _sensorLocations;
        int _dataSizeX;
        int _dataSizeY;
        double[][] _dataValues;//data values matrix
        int[][] _heatmapColorsData; //values converted into colors, and _bgImagePixels applied as well. 
        /// <summary>
        /// Helper class for computing the geometry.
        /// </summary>
        private GeometryAdditionalData[] _modelGeometryAdditionalData;
        /// <summary>
        /// Geometry centers of mass.
        /// </summary>
        private PointLineSeries3D _dataPointSeries;
        int _index = 0;
        private SeriesPoint3D[] _points;
        private int _pointCount = 0;
        // Color Palette definition.
        private Color[] _paletteStepColors = new Color[] {
            ChartTools.ColorHSVA(230),
            ChartTools.ColorHSVA(210),
            ChartTools.ColorHSVA(200),
            ChartTools.ColorHSVA(190),
            ChartTools.ColorHSVA(180),
            ChartTools.ColorHSVA(160),
            ChartTools.ColorHSVA(150),
            ChartTools.ColorHSVA(140),
            ChartTools.ColorHSVA(110),
            ChartTools.ColorHSVA(100),
            ChartTools.ColorHSVA(80),
            ChartTools.ColorHSVA(60),
            ChartTools.ColorHSVA(40),
            ChartTools.ColorHSVA(20),
            ChartTools.ColorHSVA(0)};
        private double[][] _paletteStepValues = new double[5][] {
            new double[] { 0, 1500, 2500, 4500, 5500, 6500, 8500, 10500, 12500, 13500, 18000,20000,24000,28000,35000 },
            new double[] { 0, 1000, 2000, 4000, 5000, 7000, 9000, 11000, 13000, 15000, 17000,19000,21000,23000,25000 },
            new double[] { 0, 1000, 2000, 4000, 5000, 6000, 7000, 9000, 11000, 12000, 13000,14000,15000,16000,20000 },
            new double[] { 0, 200, 400, 800, 1000, 1500, 2000, 3000, 5000, 6000, 8000,10000,12000,14000,16000 },
            new double[] { 0, 100, 300, 500, 800, 1000, 1200, 1500, 2000, 3000, 4000,5000,6000,7000,8000 }
            };

        double[][] _ThemePaletteStepValues = new double[5][];
        Color[] _ThemeStepColors;
        // Arrays with colors.
        private int[] _originalColors;
        /// <summary>
        /// Helper class for values computation from geometry data 
        /// </summary>
        public class GeometryAdditionalData
        {
            public double[] DistancesToDataPoint;
            public double[] Weights;
            public double SumOfWeights;
        }
        Thread thread = null;
        /// <summary>
        /// Calculate new data values for every center of a mass, and colors
        /// </summary>
        /// <param name="gads">Helper data</param>
        /// <returns></returns>
        private int[] CalculateGeometryColors()
        {

            // Calculate the data values for each geometry point.
            int count = _modelGeometryAdditionalData.Length;
            int[] colors = new int[count];
            double min;

            double max;
                min = _paletteStepValues[_index][0];
                max = _paletteStepValues[_index][_paletteStepValues[_index].Length - 1];
            double[] values = new double[count];

            Parallel.For(0, count, (i) =>
            {
                double sum = 0;
                GeometryAdditionalData gad = _modelGeometryAdditionalData[i];
                double[] weights = gad.Weights;
                for (int point = 0; point < _pointCount; point++)
                {
                    sum += _selected[point] * weights[point];
                }
                double value = sum / gad.SumOfWeights;
                if (value > max)
                {
                    value = max;
                }
                if (value < min)
                {
                    value = min;
                }
                values[i] = value;
            });
                return ChartTools.ConvertDataToColorsByFixedIntervalPalette(values, _paletteStepValues[_index], _paletteStepColors);

        }

        public ExampleEEGDataVisualizationShowcase()
        {
            InitializeComponent();
            InitializeCharts();
        }

        private void InitializeCharts()
        {
            CreateBuffers();
            CreateEEG();
            CreatePowerSpectrum();
            Create3DHead();
            //Create Head tops with 
            for (int i = 0; i < 5; i++)
            {
                Create2DHeads(i);
            }

            _dataSizeX = _InterpolatedResolutionX;
            _dataSizeY = _InterpolatedResolutionY;
            //Make data values array
            _dataValues = new double[_dataSizeX][];
            for (int i = 0; i < _dataSizeX; i++)
            {
                _dataValues[i] = new double[_dataSizeY];
            }

            _heatmapColorsData = new int[_dataSizeX][];
            for (int i = 0; i < _dataSizeX; i++)
            {
                _heatmapColorsData[i] = new int[_dataSizeY];
            }

            ComposeGeometryAdditionalData2D();

            thread = new Thread(new ThreadStart(AddData));
             this.Closed += MainWindow_Closed;
            foreach (LightningChart chart in _ChartList)
            {
                chart.EndUpdate();
            }
        }



        private ValueRangePalette GetPalette(double multiply, ValueRangePalette tb)
        {
            ValueRangePalette vrp = new ValueRangePalette();
            //Remove existing palette steps
            while (0 < vrp.Steps.Count)
                vrp.Steps.RemoveAndDispose<PaletteStep>(0);

            vrp.MinValue = _paletteStepValues[_index][0];
            for (int i = 0; i < tb.Steps.Count; i++)
            {
                PaletteStep ps = new PaletteStep();
                ps.Color = Color.FromArgb(255, tb.Steps[i].Color.R, tb.Steps[i].Color.G, tb.Steps[i].Color.B);
                ps.MaxValue = _paletteStepValues[_index][(int)(i * multiply >= _paletteStepValues[_index].Length ? _paletteStepValues[_index].Length - 1 : i * multiply)];
                vrp.Steps.Add(ps);
            }
            return vrp;
        }


        /// <summary>
        /// On closing dispose charts and break thread
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainWindow_Closed(object sender, EventArgs e)
        {
            Dispose();
        }

        /// <summary>
        /// Dispose charts and brake thread
        /// </summary>
        public void Dispose()
        {
            if (thread != null)
                thread.Abort();
            thread = null;
            _running = false;
            if (_ChartList == null)
                return;
            _ChartList[2].View3D.SurfaceMeshSeries3D[0].ContourPalette.Dispose();
            for (int i = 0; i < _ChartList.Count; i++)
            {
                _ChartList[i].MouseClick -= Chart_MouseClick;
                _ChartList[i].Dispose();
            }
                
            _ChartList = null;
            (ChartPlace3D.Children[0] as LightningChart).Dispose();
            ChartPlace3D.Children.Clear();
            (ChartPlaceEEG.Children[0] as LightningChart).Dispose();
            ChartPlaceEEG.Children.Clear();
            (ChartPlacePowerSpectrum.Children[0] as LightningChart).Dispose();
            ChartPlacePowerSpectrum.Children.Clear();
            (ChartHead2D0.Children[0] as LightningChart).Dispose();
            ChartHead2D0.Children.Clear();
            (ChartHead2D1.Children[0] as LightningChart).Dispose();
            ChartHead2D1.Children.Clear();
            (ChartHead2D2.Children[0] as LightningChart).Dispose();
            ChartHead2D2.Children.Clear();
            (ChartHead2D3.Children[0] as LightningChart).Dispose();
            ChartHead2D3.Children.Clear();
            (ChartHead2D4.Children[0] as LightningChart).Dispose();
            ChartHead2D4.Children.Clear();
            _reader = null;
            _spectrumCalculator.Dispose();
            _spectrumCalculator = null;

            _gads2D = null;
            _sensorLocations = null;
            _modelGeometryAdditionalData = null;
            _dataPointSeries.Dispose();
            _dataPointSeries = null;
            _points = null;
            _paletteStepColors = null;
            _ThemeStepColors = null;
            GC.Collect();
        }



        /// <summary>
        /// Update data to Charts object sender, EventArgs e
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AddData()
        {
            int count = _ChartList[0].ViewXY.SampleDataBlockSeries.Count;
            //First time
            if (_selected == null)
            {
                Dispatcher.Invoke(() =>
                {
                    _selected = _buffer04;
                        _ChartList[2].View3D.SurfaceMeshSeries3D[0].ContourPalette = GetPalette(0);
                    for (int i = 0; i < _originalColors.Length; i++)
                    {
                        _originalColors[i] = ChartTools.ColorToInt(ChartTools.ColorHSVA(1, 0, 1, 0.75));
                    }

                    _ChartList[2].View3D.MeshModels[1].UpdateFillColors(_originalColors);
                });
            }
            do
            {

                if (!_running)
                    break;
                //Begin updates
                UpdateHeads(true);

                if (_round * _iDataBufferSize > _iWinSizeForFFT - _iDataBufferSize)
                {
                    for (int i = 0; i < _EEGchannelsCount; i++)
                    {
                        Array.Copy(_buffer[i], _iDataBufferSize, _buffer[i], 0, _iWinSizeForFFT - _iDataBufferSize);
                    }
                }

                for (int i = 0; i < count; i++)
                {

                    float[] data = new float[_iDataBufferSize];
                    double average = 0;
                    for (int x = 0; x < _iDataBufferSize; x++)
                    {
                        data[x] = _reader.data[i].Item2[_round * _iDataBufferSize + x];
                        average += _reader.data[i].Item2[_round * _iDataBufferSize + x];
                    }
                    average = average / _iDataBufferSize;
                    Array.Copy(data, 0, _buffer[i], _round * _iDataBufferSize > _iWinSizeForFFT - _iDataBufferSize ? _iWinSizeForFFT - _iDataBufferSize : _round * _iDataBufferSize, data.Length);

                    // Update chart on the Main UI Thread
                    Dispatcher.Invoke(() =>
                    {
                        if (_running)
                            _ChartList[0].ViewXY.SampleDataBlockSeries[i].AddSamples(data, false);
                    }, DispatcherPriority.Send);

                    double[] fftResult;
                    bool calculated = _spectrumCalculator.PowerSpectrum(_buffer[i], out fftResult);

                    if (calculated)
                    {
                        // Fill spectrum chart.
                        int fftDataLen = fftResult.Length;
                        double xStep = _samplingFrequency / 2.0 / (double)(fftDataLen - 1);

                        // draw/render only 1-100 Hz spectrum
                        AreaSeriesPoint[] asp = new AreaSeriesPoint[(int)Math.Ceiling(_iCutOffFR / xStep)];

                        for (int iFr = 1; iFr < fftDataLen; iFr++)   // skip 0 Hz bin due to artifact
                        {
                            double dFRbin = (double)iFr * xStep;  // frequency bin
                            if (dFRbin > _iCutOffFR)       // skip Frequencies above 100 HZ
                                break;

                            asp[iFr].X = dFRbin;
                            asp[iFr].Y = fftResult[iFr];
                            if (dFRbin <= 4)
                            {
                                _buffer04[i] += fftResult[iFr];
                            }
                            if (dFRbin >= 4 && dFRbin <= 8)
                            {
                                _buffer48[i] += fftResult[iFr];
                            }
                            if (dFRbin >= 8 && dFRbin <= 12)
                            {
                                _buffer812[i] += fftResult[iFr];
                            }
                            if (dFRbin >= 12 && dFRbin <= 32)
                            {
                                _buffer1232[i] += fftResult[iFr];
                            }
                            if (dFRbin >= 32 && dFRbin <= 100)
                            {
                                _buffer32100[i] += fftResult[iFr];
                            }
                        }

                        // Update chart on the Main UI Thread
                        Dispatcher.Invoke(() =>
                        {
                            if (_running)
                                _ChartList[1].ViewXY.AreaSeries[i].Points = asp;
                        }, DispatcherPriority.Send);
                    }
                }

                for (int i = 0; i < _pointCount; i++)
                {
                    _buffer04[i] = _buffer04[i] / 4.0;
                    _buffer48[i] = _buffer48[i] / 5.0;
                    _buffer812[i] = _buffer812[i] / 5.0;
                    _buffer1232[i] = _buffer1232[i] / 21.0;
                    _buffer32100[i] = _buffer32100[i] / 79.0;
                }

                int sensorCount = _sensorLocations.Length; // _dataPointValues.Length;
                int[] colors;
                    colors = ChartTools.ConvertDataToColorsByFixedIntervalPalette(_selected, _paletteStepValues[_index], _paletteStepColors);

                // Update chart on the Main UI Thread
                Dispatcher.Invoke(() =>
                {
                    if (_running)
                        for (int i = 0; i < sensorCount; i++)
                        {
                            _points[i].Color = FromArgb(colors[i]);
                        }
                }, DispatcherPriority.Send);

                UpdateHeadHeatmaps(_buffer04, 0);
                UpdateHeadHeatmaps(_buffer48, 1);
                UpdateHeadHeatmaps(_buffer812, 2);
                UpdateHeadHeatmaps(_buffer1232, 3);
                UpdateHeadHeatmaps(_buffer32100, 4);

                int[] aBrainColors = CalculateGeometryColors();

                // Update chart on the Main UI Thread
                Dispatcher.Invoke(() =>
                {
                    if (_running)
                    {
                        _ChartList[2].View3D.PointLineSeries3D[0].InvalidateData();
                        _ChartList[2].View3D.MeshModels[0].UpdateFillColors(aBrainColors);

                        _ChartList[0].ViewXY.XAxes[0].ScrollPosition = _ChartList[0].ViewXY.SampleDataBlockSeries[0].PointCount / _samplingFrequency;
                    }
                }, DispatcherPriority.Send);

                UpdateHeads(false);

                _round++;
                if (_round * _iDataBufferSize + _iDataBufferSize > _reader.data[0].Item2.Count)
                {
                    _round = 0;
                }

            } while (_running);
        }

        private void UpdateHeads(bool begin)
        {
            if (begin)
            {
                if (_running)
                {
                    for (int i = 0; i < _ChartList.Count; i++)
                    {
                        _ChartList[i].BeginUpdate();
                    }
                }
            }
            else
            {
                try
                {
                    for (int i = 0; i < _ChartList.Count; i++)
                    {
                        if (_running)
                        {
                            Dispatcher.Invoke(() =>
                            {
                                if(_ChartList != null)
                                    _ChartList[i].EndUpdate();
                            });
                        }
                    }
                }
                catch { }
            }
        }

        #region Crate Charts

        private void Create2DHeads(int place)
        {
            LightningChart chart = new LightningChart();
            chart.BeginUpdate();
            confchart(chart);
            chart.ActiveView = ActiveView.ViewXY;
            GetTheme(chart);
            switch (place)
            {
                case 0:
                    chart.Title.Text = "Delta P(f) f=0-4Hz";
                    chart.ChartName = "DeltaHead";
                    chart.MouseClick += Chart_MouseClick;
                    ChartHead2D0.Children.Add(chart);
                    break;
                case 1:
                    chart.Title.Text = "Theta P(f) f=4-8Hz";
                    chart.ChartName = "ThetaHead";
                    chart.MouseClick += Chart_MouseClick;
                    ChartHead2D1.Children.Add(chart);
                    break;
                case 2:
                    chart.Title.Text = "Alpha P(f) f=8-12Hz";
                    chart.ChartName = "AlphaHead";
                    chart.MouseClick += Chart_MouseClick;
                    ChartHead2D2.Children.Add(chart);
                    break;
                case 3:
                    chart.Title.Text = "Beta P(f) f=12-32Hz";
                    chart.ChartName = "BetaHead";
                    chart.MouseClick += Chart_MouseClick;
                    ChartHead2D3.Children.Add(chart);
                    break;
                case 4:
                    chart.Title.Text = "Gamma P(f) f=32-100Hz";
                    chart.ChartName = "GammaHead";
                    chart.MouseClick += Chart_MouseClick;
                    ChartHead2D4.Children.Add(chart);
                    break;
            }
            chart.ViewXY.YAxes.Add(new AxisY());
            chart.Title.Align = ChartTitleAlignment.TopCenter;
            chart.Title.AllowUserInteraction = false;
            chart.ViewXY.XAxes[0].MinorGrid.Visible = false;
            chart.ViewXY.XAxes[0].Title.AllowUserInteraction = false;
            chart.ViewXY.XAxes[0].MajorGrid.Visible = false;
            chart.ViewXY.XAxes[0].SetRange(-110, 110);
            chart.ViewXY.YAxes[0].SetRange(-100, 100);
            chart.ViewXY.YAxes[1].SetRange(-85, 105);
            chart.ViewXY.YAxes[1].MinorGrid.Visible = false;
            chart.ViewXY.YAxes[1].MajorGrid.Visible = false;
            chart.ViewXY.YAxes[1].Visible = false;
            chart.ViewXY.YAxes[0].MinorGrid.Visible = false;
            chart.ViewXY.YAxes[0].MajorGrid.Visible = false;
            chart.ViewXY.YAxes[0].Title.AllowUserInteraction = false;
            chart.ViewXY.YAxes[1].Title.AllowUserInteraction = false;
            chart.ViewXY.YAxes[0].Visible = false;
            chart.ViewXY.XAxes[0].Visible = false;
            chart.ViewXY.ZoomPanOptions.WheelZooming = WheelZooming.Off;
            chart.ViewXY.ZoomPanOptions.RightToLeftZoomAction = RightToLeftZoomActionXY.Off;
            chart.ViewXY.ZoomPanOptions.DevicePrimaryButtonAction = UserInteractiveDeviceButtonAction.None;
            Assembly assembly = Assembly.GetExecutingAssembly();
            System.Drawing.Bitmap bmp = (System.Drawing.Bitmap)ChartTools.ImageFromResource("Content.Head.png", assembly);
            //Make image in the background 
            IntensityGridSeries _bgImage;
            _bgImage = new IntensityGridSeries(chart.ViewXY, chart.ViewXY.XAxes[0], chart.ViewXY.YAxes[0]);
            _bgImage.PixelRendering = true;
            _bgImage.SetHeightDataFromBitmap(_XMin - 0.5, _XMax - 0.5, _YMin, _YMax, 0, 100, bmp, new BitmapAntialiasOptions());
            _bgImage.Fill = IntensityFillStyle.FromSurfacePoints;
            _bgImage.ContourLineType = ContourLineTypeXY.None;
            _bgImage.ContourLineStyle.Color = Colors.Red;
            _bgImage.WireframeType = SurfaceWireframeType.None;
            _bgImage.Title.Text = "BG image";
            chart.ViewXY.IntensityGridSeries.Add(_bgImage);
            PointDouble2D[] stencilPath = ReadStencil(assembly, "headContour.csv");
            IntensityGridSeries _heatmap;
            _heatmap = new IntensityGridSeries(chart.ViewXY, chart.ViewXY.XAxes[0], chart.ViewXY.YAxes[1]);
            _heatmap.PixelRendering = true;
            _heatmap.Optimization = IntensitySeriesOptimization.DynamicValuesData;
            _heatmap.ContourLineType = ContourLineTypeXY.None;
            _heatmap.WireframeType = SurfaceWireframeType.None;
            _heatmap.SetRangesXY(_XMin, _XMax, _YMin, _YMax);
            _heatmap.Fill = IntensityFillStyle.FromSurfacePoints;
            _heatmap.SetSize(_InterpolatedResolutionX, _InterpolatedResolutionY);
            chart.ViewXY.IntensityGridSeries.Add(_heatmap);
            chart.CustomDynamicTheme.DynamicBackgroundFillStyle = RectFillStyle.ColorOnly;
            chart.UpdateCustomTheme();
            chart.ChartBackground.Color = ChartTools.ColorHSVA(place * 48, 0.5, 0.4, 0.3);
            //Apply stencil for the heatmap. Especially the Interpolated coloring needs this, as it can't be combined with the alpha mask of the bitmap. 
            StencilArea stencilArea = new StencilArea(_heatmap.Stencil);
            stencilArea.AddPolygon(stencilPath);
            _heatmap.Stencil.AdditiveAreas.Add(stencilArea);
            _sensorLocations = MakeEEG2D_Locations();
            CreateActiveSensorSymbols(chart.ViewXY, chart.ViewXY.XAxes[0], chart.ViewXY.YAxes[0]);

            _ChartList.Add(chart);

        }
        private void Create3DHead()
        {
            LightningChart chart = new LightningChart();
            chart.BeginUpdate();
            chart.View3D.Camera.Projection = ProjectionType.Orthographic;

            confchart(chart);
            GetTheme(chart);
            chart.ChartName = "3DHead";
            chart.Title.Text = "Delta P(f) f=0-4Hz";
            chart.Title.Align = ChartTitleAlignment.BottomCenter;
            chart.View3D.XAxisPrimary3D.Visible = false;
            chart.View3D.YAxisPrimary3D.Visible = false;
            chart.View3D.ZAxisPrimary3D.Visible = false;
            chart.ActiveView = ActiveView.View3D;
            chart.View3D.Camera.ViewDistance = 300;
            chart.View3D.Dimensions = new SizeDoubleXYZ(200, 200, 200);
            chart.View3D.ZoomPanOptions.AllowWheelZoom = false;
            foreach (WallBase wall in chart.View3D.GetWalls())
            {
                wall.Visible = false;
            }

            MeshModel MeashBrains = new MeshModel();
            MeashBrains.LoadFromFile(_Brains);
            MeashBrains.AllowUserInteraction = false;
            MeashBrains.Position = new PointDoubleXYZ(50, 45 /*-360*/, 41);
            MeashBrains.Size.Width = 70;
            MeashBrains.Size.Depth = 70;
            MeashBrains.Size.Height = 70;

            MeshModel MeshHead = new MeshModel();
            MeshHead.LoadFromFile(_Head);
            MeshHead.AllowUserInteraction = false;
            MeshHead.Position = new PointDoubleXYZ(50, -360, 50);
            MeshHead.Size.Width = 500;
            MeshHead.Size.Depth = 500;
            MeshHead.Size.Height = 500;

            MeshHead.GeometryConstructed += head_GeometryConstructed;
            MeashBrains.GeometryConstructed += brain_GeometryConstructed;
            // Create data.
            _points = MakeDataPoints();
            _pointCount = _points.Length;

            // Create series to show geometry center points.
            _dataPointSeries = new PointLineSeries3D(chart.View3D, Axis3DBinding.Primary, Axis3DBinding.Primary, Axis3DBinding.Primary)
            {
                Points = _points
            };
            _dataPointSeries.PointStyle.Shape3D = PointShape3D.Sphere;
            _dataPointSeries.LineVisible = false;
            _dataPointSeries.IndividualPointColors = true;
            _dataPointSeries.AllowUserInteraction = true;
            _dataPointSeries.PointStyle.Size3D.SetValues(2, 2, 2);
            _dataPointSeries.Visible = true;
            _dataPointSeries.Highlight = Highlight.None;
            _dataPointSeries.PointStyle.Shape2D.Shape = Shape.Circle;
            _dataPointSeries.PointStyle.Shape2D.GradientFill = GradientFillPoint.Solid;
            _dataPointSeries.PointStyle.Shape2D.Width = 50;
            _dataPointSeries.PointStyle.Shape2D.Height = 50;
            _dataPointSeries.ShowInLegendBox = false;
            chart.View3D.PointLineSeries3D.Add(_dataPointSeries);
            chart.View3D.MeshModels.Add(MeashBrains);
            chart.View3D.MeshModels.Add(MeshHead);
            SurfaceMeshSeries3D dummy = new SurfaceMeshSeries3D();
            dummy.WireframeType = SurfaceWireframeType3D.None;
            dummy.ContourLineType = ContourLineType3D.None;
            dummy.Title.Text = "Activity";
            dummy.Title.Visible = false;
            chart.View3D.SurfaceMeshSeries3D.Add(dummy);
            chart.CustomDynamicTheme.DynamicBackgroundFillStyle = RectFillStyle.None;
            chart.UpdateCustomTheme();
            chart.View3D.LegendBox.Visible = true;
            chart.View3D.LegendBox.ShowCheckboxes = false;
            chart.View3D.LegendBox.ShowIcons = false;
            chart.View3D.LegendBox.AllowUserInteraction = false;
            chart.View3D.LegendBox.Shadow.Visible = false;
            chart.View3D.YAxisPrimary3D.Units.Text = "μV^2/Hz";
            chart.View3D.LegendBox.BorderWidth = 0;
            chart.Title.AllowUserInteraction = false;
            chart.View3D.LegendBox.ScrollBarVisibility = Arction.Wpf.Charting.Views.LegendBoxBase.LegendBoxScrollBarVisibility.None;
            chart.SizeChanged += Chart_SizeChanged;
            Annotation3D anno = new Annotation3D();
            anno.AllowDragging = false;
            anno.MouseClick += Anno_MouseClick;
            anno.Text = "Hide head model";
            anno.Style = AnnotationStyle.RoundedRectangle;
            anno.AllowResize = false;
            anno.BorderVisible = false;
            anno.Fill.Color = Colors.OrangeRed;
            anno.TextStyle.Color = Colors.Black;
            anno.AllowRotate = false;
            anno.AllowTargetMove = false;
            anno.AllowAnchorAdjust = false;
            anno.ShowTargetCrosshair = TargetCrossHair.Off;
            chart.View3D.Annotations.Add(anno);
            ChartPlace3D.Children.Add(chart);
            _ChartList.Add(chart);
        }

        private void Anno_MouseClick(object sender, MouseEventArgs e)
        {
            Annotation3D anno = sender as Annotation3D;
            if (anno.Text == "Hide head model")
            {
                anno.Text = "Reveal head model";
                _ChartList[2].View3D.MeshModels[1].Visible = false;
            }
            else
            {
                anno.Text = "Hide head model";
                _ChartList[2].View3D.MeshModels[1].Visible = true;
            }
        }

        private void Chart_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            _ChartList[2].View3D.LegendBox.SurfaceScales.ScaleSizeDim1 = (int)e.NewSize.Height - 100;
        }

        private void CreatePowerSpectrum()
        {
            LightningChart chart = new LightningChart();
            _spectrumCalculator = new SpectrumCalculator();
            chart.BeginUpdate();
            chart.ChartName = "PowerSpectrum";
            confchart(chart);
            GetTheme(chart);
            chart.ActiveView = ActiveView.ViewXY;
            chart.ViewXY.XAxes[0].MinorGrid.Visible = false;
            chart.ViewXY.XAxes[0].MajorGrid.Visible = false;
            chart.ViewXY.XAxes[0].Title.AllowUserInteraction = false;
            chart.ViewXY.XAxes[0].AllowUserInteraction = false;
            chart.ViewXY.XAxes[0].Title.Text = "P(f) f = 0-100Hz";
            chart.ViewXY.ZoomPanOptions.WheelZooming = WheelZooming.Off;
            chart.ViewXY.ZoomPanOptions.RightToLeftZoomAction = RightToLeftZoomActionXY.Off;
            chart.ViewXY.ZoomPanOptions.DevicePrimaryButtonAction = UserInteractiveDeviceButtonAction.None;
            chart.ViewXY.YAxes.Clear();
            for (int i = 0; i < _reader.data.Count; i++)
            {
                chart.ViewXY.YAxes.Add(new AxisY());
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Minimum = 0;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Maximum = 200000;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Title.Text = _reader.data[i].Item1;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].MinorGrid.Visible = false;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].MajorGrid.Visible = false;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].ScaleNibs.Size = new SizeDoubleXY(0, 0);
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].AllowUserInteraction = false;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Title.AllowUserInteraction = false;
                chart.ViewXY.AreaSeries.Add(new AreaSeries());
                chart.ViewXY.AreaSeries[i].AssignYAxisIndex = i;
                chart.ViewXY.AreaSeries[i].AllowUserInteraction = false;
                chart.ViewXY.AreaSeries[i].Fill.GradientFill = GradientFill.Solid;
                chart.ViewXY.AreaSeries[i].Fill.Color = ChartTools.ColorHSVA(i * 11, 1, 1, 0.5);
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].LabelsColor = chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].AxisColor
                    = chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Title.Color
                    = chart.ViewXY.AreaSeries[i].LineStyle.Color = ChartTools.ColorHSVA(i * 11);
            }
            for (int i = 0; i < 5; i++)
            {
                chart.ViewXY.Bands.Add(NewBand(i));
            }
            AnnotationXY anno = new AnnotationXY();
            anno.AllowDragging = false;

            anno.Style = AnnotationStyle.RoundedRectangle;
            anno.AllowResize = false;
            anno.BorderVisible = false;
            anno.Fill.Color = Colors.OrangeRed;
            anno.TextStyle.Color = Colors.Black;
            anno.AllowRotate = false;
            anno.AllowTargetMove = false;
            anno.AllowAnchorAdjust = false;
            anno.Visible = false;
            anno.AllowUserInteraction = false;
            anno.ClipInsideGraph = false;
            chart.ViewXY.Annotations.Add(anno);
            chart.ViewXY.XAxes[0].Minimum = 0;
            chart.ViewXY.XAxes[0].Maximum = _iCutOffFR;
            chart.ViewXY.AxisLayout.YAxesLayout = YAxesLayout.Stacked;
            chart.MouseMove += Chart_MouseMove;
            ChartPlacePowerSpectrum.Children.Add(chart);
            _ChartList.Add(chart);
        }

        private void Chart_MouseMove(object sender, MouseEventArgs e)
        {
            LightningChart lc = sender as LightningChart;
            Point mousepoint = e.GetPosition(lc);


            if (lc.ViewXY.Bands[0].IsPositionOver((int)mousepoint.X, (int)mousepoint.Y, true))
            {
                lc.ViewXY.Annotations[0].Text = "Delta 0-4Hz";
                lc.ViewXY.Annotations[0].Visible = true;
                lc.ViewXY.Annotations[0].LocationScreenCoords = AnnotationPosition(mousepoint, lc);
            }
            else if (lc.ViewXY.Bands[1].IsPositionOver((int)mousepoint.X, (int)mousepoint.Y, true))
            {
                lc.ViewXY.Annotations[0].Text = "Theta 4-8Hz";
                lc.ViewXY.Annotations[0].Visible = true;
                lc.ViewXY.Annotations[0].LocationScreenCoords = AnnotationPosition(mousepoint, lc);
            }
            else if (lc.ViewXY.Bands[2].IsPositionOver((int)mousepoint.X, (int)mousepoint.Y, true))
            {
                lc.ViewXY.Annotations[0].Text = "Alpha 8-12Hz";
                lc.ViewXY.Annotations[0].Visible = true;
                lc.ViewXY.Annotations[0].LocationScreenCoords = AnnotationPosition(mousepoint, lc);
            }
            else if (lc.ViewXY.Bands[3].IsPositionOver((int)mousepoint.X, (int)mousepoint.Y, true))
            {
                lc.ViewXY.Annotations[0].Text = "Beta 12-32Hz";
                lc.ViewXY.Annotations[0].Visible = true;
                lc.ViewXY.Annotations[0].LocationScreenCoords = AnnotationPosition(mousepoint, lc);
            }
            else if (lc.ViewXY.Bands[4].IsPositionOver((int)mousepoint.X, (int)mousepoint.Y, true))
            {
                lc.ViewXY.Annotations[0].Text = "Gamma 32-100Hz";
                lc.ViewXY.Annotations[0].Visible = true;
                lc.ViewXY.Annotations[0].LocationScreenCoords = AnnotationPosition(mousepoint, lc);
            }
            else
            {
                lc.ViewXY.Annotations[0].Visible = false;
            }
        }

        private PointDoubleXY AnnotationPosition(Point mousepointDIP, LightningChart lc)
        {
            Point mousepoint = DpiHelper.DipToPx(mousepointDIP);
            PointDoubleXY pd = new PointDoubleXY();

            if (mousepoint.X + lc.ViewXY.Annotations[0].SizeScreenCoords.Width / 2 > lc.ActualWidth - lc.ViewXY.Margins.Left)
            {
                pd.X = mousepoint.X - lc.ViewXY.Annotations[0].SizeScreenCoords.Width / 2;
            }
            else
            {
                pd.X = mousepoint.X + lc.ViewXY.Annotations[0].SizeScreenCoords.Width / 2;
            }
            if (mousepoint.Y + lc.ViewXY.Annotations[0].SizeScreenCoords.Height / 2 > lc.ActualHeight - lc.ViewXY.Margins.Bottom)
            {
                pd.Y = mousepoint.Y - lc.ViewXY.Annotations[0].SizeScreenCoords.Height / 2;
            }
            else
            {
                pd.Y = mousepoint.Y + lc.ViewXY.Annotations[0].SizeScreenCoords.Height / 2;
            }
            return pd;
        }

        /// <summary>
        /// Create band for power spectrum to indicate areas
        /// </summary>
        /// <param name="i"></param>
        /// <returns></returns>
        private Band NewBand(int i)
        {
            Band band = new Band();
            band.Behind = true;
            band.Fill.GradientFill = GradientFill.Solid;
            band.Fill.Color = ChartTools.ColorHSVA(i * 48, 0.5, 0.4, 0.3);
            band.Fill.Style = RectFillStyle.ColorOnly;
            band.UseLimits = true;
            band.AllowUserInteraction = false;

            if (i == 0)
            {
                band.Title.Text = "Delta";
                band.ValueBegin = 0;
                band.ValueEnd = 4;
            }
            else if (i == 1)
            {
                band.Title.Text = "Theta";
                band.ValueBegin = 4;
                band.ValueEnd = 8;
            }
            else if (i == 2)
            {
                band.Title.Text = "Alpha";
                band.ValueBegin = 8;
                band.ValueEnd = 16;
            }
            else if (i == 3)
            {
                band.Title.Text = "Beta";
                band.ValueBegin = 16;
                band.ValueEnd = 32;
            }
            else if (i == 4)
            {
                band.Title.Text = "Gamma";
                band.ValueBegin = 32;
                band.ValueEnd = 100;
            }
            return band;
        }

        private void CreateEEG()
        {
            ChartManager cm = new ChartManager();
            LightningChart chart = new LightningChart();
            chart.BeginUpdate();
            chart.ChartName = "EEG";
            confchart(chart);
            GetTheme(chart);
            chart.ViewXY.ZoomPanOptions.WheelZooming = WheelZooming.Off;
            chart.ViewXY.ZoomPanOptions.RightToLeftZoomAction = RightToLeftZoomActionXY.Off;
            chart.ViewXY.ZoomPanOptions.DevicePrimaryButtonAction = UserInteractiveDeviceButtonAction.None;
            chart.ActiveView = ActiveView.ViewXY;
            chart.ViewXY.XAxes[0].MinorGrid.Visible = false;
            chart.ViewXY.XAxes[0].MajorGrid.Visible = false;
            chart.ViewXY.XAxes[0].AllowUserInteraction = false;
            chart.ViewXY.XAxes[0].AutoDivSpacing = false;
            chart.ViewXY.XAxes[0].MajorDiv = 10;
            chart.ViewXY.XAxes[0].Title.Text = "Time (s)";
            chart.ViewXY.AxisLayout.AutoAdjustMargins = false;
            chart.ViewXY.Margins = new Thickness(65, 7, 10, 50);
            chart.ViewXY.YAxes.Clear();
            for (int i = 0; i < _reader.data.Count; i++)
            {
                chart.ViewXY.YAxes.Add(new AxisY());
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Minimum = -2000;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Maximum = 2000;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Title.Text = _reader.data[i].Item1;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].MinorGrid.Visible = false;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].MajorGrid.Visible = false;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].AllowUserInteraction = false;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].ScaleNibs.Size = new SizeDoubleXY(0, 0);
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Title.AllowUserInteraction = false;
                chart.ViewXY.SampleDataBlockSeries.Add(new SampleDataBlockSeries());
                chart.ViewXY.SampleDataBlockSeries[i].AllowUserInteraction = false;
                chart.ViewXY.SampleDataBlockSeries[i].AssignYAxisIndex = i;
                chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].LabelsColor = chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].AxisColor
                    = chart.ViewXY.YAxes[chart.ViewXY.YAxes.Count - 1].Title.Color
                    = chart.ViewXY.SampleDataBlockSeries[i].Color = ChartTools.ColorHSVA(i * 11);
                chart.ViewXY.SampleDataBlockSeries[i].SamplingFrequency = _samplingFrequency;
                chart.ViewXY.SampleDataBlockSeries[i].FirstSampleTimeStamp = 1 / _samplingFrequency;
            }
            chart.ViewXY.AxisLayout.YAxesLayout = YAxesLayout.Stacked;
            chart.ViewXY.XAxes[0].Minimum = 0;
            chart.ViewXY.XAxes[0].Maximum = 30;     // 30 sec
            chart.ViewXY.XAxes[0].ScrollMode = XAxisScrollMode.Scrolling;
            ChartPlaceEEG.Children.Add(chart);
            _ChartList.Add(chart);
        }
        #endregion

        /// <summary>
        /// Create active sensor symbols
        /// </summary>
        private void CreateActiveSensorSymbols(ViewXY v, AxisX _xAxis, AxisY _yAxis)
        {
            FreeformPointLineSeries seriesSensorPositions;
            //Add series to represent sensors. Positions, values and colors are applied. 


            seriesSensorPositions = new FreeformPointLineSeries(v, _xAxis, _yAxis);
            seriesSensorPositions.LineVisible = false;
            seriesSensorPositions.AllowUserInteraction = true;
            seriesSensorPositions.Visible = true;
            seriesSensorPositions.Highlight = Highlight.None;
            seriesSensorPositions.PointStyle.Shape = Arction.Wpf.Charting.Shape.Circle;
            seriesSensorPositions.PointStyle.GradientFill = GradientFillPoint.Solid;
            seriesSensorPositions.PointStyle.Width = 11;
            seriesSensorPositions.PointStyle.Height = 11;
            seriesSensorPositions.ShowInLegendBox = false;

            v.FreeformPointLineSeries.Add(seriesSensorPositions);

            List<SeriesPoint> activeLocations = new List<SeriesPoint>();
            seriesSensorPositions.Points = activeLocations.ToArray();

            MakeMarkersForSensors(seriesSensorPositions);
        }
        /// <summary>
        /// Make markers for sensors 
        /// </summary>
        void MakeMarkersForSensors(FreeformPointLineSeries seriesSensorPositions)
        {

            int activeSensorsCount = _sensorLocations.Length;

            for (int i = 0; i < activeSensorsCount; i++)
            {
                SeriesPoint p = _sensorLocations[i];

                SeriesEventMarker marker = new SeriesEventMarker(seriesSensorPositions);
                marker.Symbol.Color1 = Colors.Gray;
                marker.Symbol.GradientFill = GradientFillPoint.Solid;
                marker.VerticalPosition = SeriesEventMarkerVerticalPosition.AtYValue;
                marker.HorizontalPosition = SeriesEventMarkerHorizontalPosition.AtXValue;
                marker.AllowUserInteraction = false;
                marker.AllowDragging = true;
                marker.Symbol.Shape = SeriesMarkerPointShape.Circle;
                marker.Symbol.Width = 4;
                marker.Symbol.Height = 4;
                marker.Symbol.BorderWidth = 0;
                marker.XValue = _sensorLocations[i].X;
                marker.YValue = _sensorLocations[i].Y;
                marker.Label.Text = p.Tag.ToString();
                marker.Label.HorizontalAlign = AlignmentHorizontal.Center;
                marker.Label.VerticalAlign = AlignmentVertical.Top;
                seriesSensorPositions.SeriesEventMarkers.Add(marker);
            }
        }

        PointDouble2D[] ReadStencil(Assembly assembly, string resourceName)
        {

            //Read headContour as stencil path. 
            string resourceStream = assembly.GetName().Name + "." + "Content." + resourceName;

            Boolean bOk = true;
            List<PointDouble2D> listPoints = new List<PointDouble2D>();

            try
            {
                using (System.IO.Stream s = assembly.GetManifestResourceStream(resourceStream))
                {
                    using (StreamReader sr = new StreamReader(s))
                    {
                        Char cValueSepar = ',';

                        //ChartTools.GetCSVSeparatorChars(separator, out cValueSepar, out cDecimalSepar);

                        String strLine = sr.ReadLine(); // Skip the header.
                        strLine = sr.ReadLine();

                        while (strLine != null)
                        {
                            String[] aValues = strLine.Split(cValueSepar);

                            if (aValues.Length != 2)
                            {
                                bOk = false;

                                break;
                            }

                            PointDouble2D point = new PointDouble2D();
                            point.X = Double.Parse(aValues[0], System.Globalization.CultureInfo.InvariantCulture);
                            point.Y = Double.Parse(aValues[1], System.Globalization.CultureInfo.InvariantCulture);
                            listPoints.Add(point);
                            strLine = sr.ReadLine();
                        }
                    }
                }
            }
            catch
            {
                bOk = false;
            }
            if (bOk)
            {
                listPoints.Add(new PointDouble2D(listPoints[0].X, listPoints[0].Y));
                //listPoints.Reverse();
                return listPoints.ToArray();
            }
            return null;
        }
        private void Chart_MouseClick(object sender, MouseButtonEventArgs e)
        {
            switch ((sender as LightningChart).Title.Text)
            {
                case "Delta P(f) f=0-4Hz":
                    _selected = _buffer04;
                    _ChartList[2].Title.Text = "Delta P(f) f = 0 - 4Hz";
                    _index = 0;
                    _ChartList[2].View3D.SurfaceMeshSeries3D[0].ContourPalette = GetPalette(0);
                    break;
                case "Theta P(f) f=4-8Hz":
                    _selected = _buffer48;
                    _ChartList[2].Title.Text = "Theta P(f) f = 4 - 8Hz";
                    _index = 1;
                    _ChartList[2].View3D.SurfaceMeshSeries3D[0].ContourPalette = GetPalette(1);
                    break;
                case "Alpha P(f) f=8-12Hz":
                    _selected = _buffer812;
                    _ChartList[2].Title.Text = "Alpha P(f) f = 8 - 12Hz";
                    _index = 2;
                    _ChartList[2].View3D.SurfaceMeshSeries3D[0].ContourPalette = GetPalette(2);
                    break;
                case "Beta P(f) f=12-32Hz":
                    _selected = _buffer1232;
                    _ChartList[2].Title.Text = "Beta P(f) f = 12 - 32Hz";
                    _index = 3;
                    _ChartList[2].View3D.SurfaceMeshSeries3D[0].ContourPalette = GetPalette(3);
                    break;
                case "Gamma P(f) f=32-100Hz":
                    _selected = _buffer32100;
                    _ChartList[2].Title.Text = "Gamma P(f) f = 32 - 100Hz";
                    _index = 4;
                    _ChartList[2].View3D.SurfaceMeshSeries3D[0].ContourPalette = GetPalette(4);
                    break;
            }
        }

        private ValueRangePalette GetPalette(int v)
        {
            ValueRangePalette vrp = new ValueRangePalette();
            //Remove existing palette steps
            while (0 < vrp.Steps.Count)
                vrp.Steps.RemoveAndDispose<PaletteStep>(0);

                vrp.MinValue = _paletteStepValues[v][0];
                for (int i = 0; i < _paletteStepValues[v].Length; i++)
                {
                    PaletteStep ps = new PaletteStep();
                    ps.Color = _paletteStepColors[i];
                    ps.MaxValue = _paletteStepValues[v][i];
                    vrp.Steps.Add(ps);
                }
            return vrp;
        }

        private void confchart(LightningChart chart)
        {
            chart.Title.Text = "";
            chart.ViewXY.LegendBoxes[0].Visible = false;
            chart.View3D.LegendBox.Visible = false;
        }

        /// <summary>
        /// Create 3D data points in view
        /// </summary>
        /// <returns></returns>
        private SeriesPoint3D[] MakeDataPoints()
        {
            SeriesPoint3D[] points = new SeriesPoint3D[] { 
                 // AF3
                 new SeriesPoint3D(59.95, 68.40, 18.33),
                 // AF4
                 new SeriesPoint3D(40.19, 68.40, 18.33),
                 // F5
                 new SeriesPoint3D(63.47, 75.46, 26.09),
                 // Fz
                 new SeriesPoint3D(50, 82.51, 25.85),
                 // F6
                 new SeriesPoint3D(37.37, 75.46, 26.09),
                 // FC5
                 new SeriesPoint3D(64.88, 77.57, 34.56),
                 // FCZ
                 new SeriesPoint3D(50, 85.33, 34.56),
                 // FC6
                 new SeriesPoint3D(35.26, 77.57, 34.56),
                 // T7
                 new SeriesPoint3D(68.97, 68.41, 44.43),
                 // C3
                 new SeriesPoint3D(63.33, 80.39, 45.14),
                 // Cz
                 new SeriesPoint3D(50, 85.75, 45.14), 
                 // C4
                 new SeriesPoint3D(36.67, 80.39, 45.14),
                 // T8
                 new SeriesPoint3D(31.03, 68.41, 44.43),
                 // CP5
                 new SeriesPoint3D(64.88, 75.46, 54.30),
                 // CPZ
                 new SeriesPoint3D(50, 80.39, 58.53),
                 // CP6
                 new SeriesPoint3D(35.26, 75.46, 54.30),
                 // P5
                 new SeriesPoint3D(62.76, 67.70, 61.53),
                 // Pz
                 new SeriesPoint3D(50, 69.11, 65.70),
                 // P6
                 new SeriesPoint3D(37.37, 67.70, 61.53),
                 // O1
                 new SeriesPoint3D(44.74, 55.00, 64.76),
                 // O2
                 new SeriesPoint3D(56.04, 55.00, 64.76)
            };

            return points;
        }
        private void brain_GeometryConstructed(object sender, GeometryConstructedEventArgs e)
        {
            // Make helper data to help geometry and colors computation in real-time. 
            _modelGeometryAdditionalData = ComposeGeometryAdditionalData(e.X, e.Y, e.Z);

            // Unsubscribe from the event. 
            e.Mesh.GeometryConstructed -= brain_GeometryConstructed;
        }

        private void head_GeometryConstructed(object sender, GeometryConstructedEventArgs e)
        {
            _originalColors = new int[e.X.Length];

            // Unsubscribe from the event. 
            e.Mesh.GeometryConstructed -= head_GeometryConstructed;
            thread.Start();
        }

        /// <summary>
        /// Make helper data to help geometry and colors computation in real-time 
        /// </summary>
        /// <param name="x">World coordinate x</param>
        /// <param name="y">World coordinate y</param>
        /// <param name="z">World coordinate z</param>
        /// <returns>Helper Data</returns>
        private GeometryAdditionalData[] ComposeGeometryAdditionalData(float[] x, float[] y, float[] z)
        {
            AxisX3D xAxis = _ChartList[2].View3D.XAxisPrimary3D;
            AxisY3D yAxis = _ChartList[2].View3D.YAxisPrimary3D;
            AxisZ3D zAxis = _ChartList[2].View3D.ZAxisPrimary3D;

            int count = x.Length;
            GeometryAdditionalData[] gads = new GeometryAdditionalData[count];


            System.Threading.Tasks.Parallel.For(0, count, (i) =>
            {
                double xValue = xAxis.World3DCoordToValue(x[i]);
                double yValue = yAxis.World3DCoordToValue(y[i]);
                double zValue = zAxis.World3DCoordToValue(z[i]);

                GeometryAdditionalData gad = new GeometryAdditionalData
                {
                    DistancesToDataPoint = new double[_pointCount],
                    Weights = new double[_pointCount]
                };

                // Use weighted average method.
                double sumOfWeights = 0;
                for (int point = 0; point < _pointCount; point++)
                {
                    double distance = CalcDist(xValue, yValue, zValue, point);
                    gad.DistancesToDataPoint[point] = distance;

                    double weight = 1;
                    if (distance != 0)
                    {
                        weight = 1.0 / Math.Pow(distance, 3);
                    }
                    gad.Weights[point] = weight;

                    sumOfWeights += weight;
                }

                gad.SumOfWeights = sumOfWeights;

                gads[i] = gad;

            });
            return gads;
        }

        /// <summary>
        /// Calculate distance between 3D points
        /// </summary>
        /// <param name="x">X value</param>
        /// <param name="y">Y value</param>
        /// <param name="z">Z value</param>
        /// <param name="pointIndex">Point index in the array</param>
        /// <returns>Distance</returns>
        private double CalcDist(double x, double y, double z, int pointIndex)
        {
            SeriesPoint3D p = _points[pointIndex];
            double dist = Math.Sqrt((x - p.X) * (x - p.X) + (y - p.Y) * (y - p.Y) + (z - p.Z) * (z - p.Z));
            return dist;
        }



        private void GetTheme(LightningChart chart)
        {
            chart.ColorTheme = ColorTheme.CustomDynamicTheme;
            LightningChart.ThemeBasics tb = chart.GetThemeAuroras();
            tb.DynamicBackgroundLayout = BitmapFillLayout.Fit;
            tb.BackgroundColor = Colors.Black;
            tb.BackgroundGradientColor = Colors.Black;
            tb.GraphBackgroundGradientColor = Colors.Black;
            tb.DynamicBackgroundFillStyle = RectFillStyle.ColorOnly;
            tb.GraphBackgroundColor = Colors.Black;
            chart.CustomDynamicTheme = tb;
            chart.UpdateCustomTheme();
        }




        /// <summary>
        /// 2D heads EEG points locations
        /// </summary>
        /// <returns></returns>
        SeriesPoint[] MakeEEG2D_Locations()
        {
            List<SeriesPoint> listPoints = new List<SeriesPoint>();
            //AF3
            SeriesPoint AF3 = new SeriesPoint(-23.6326, 60.551);
            AF3.Tag = "AF3";
            listPoints.Add(AF3);

            //AF4 
            SeriesPoint AF4 = new SeriesPoint(19.2945, 60.511);
            AF4.Tag = "AF4";
            listPoints.Add(AF4);

            //F5 
            SeriesPoint F5 = new SeriesPoint(-32.7989, 38.944);
            F5.Tag = "F5";
            listPoints.Add(F5);

            //Fz
            SeriesPoint Fz = new SeriesPoint(0, 38.944);
            Fz.Tag = "Fz";
            listPoints.Add(Fz);

            //F6 
            SeriesPoint F6 = new SeriesPoint(30.146, 38.944);
            F6.Tag = "F6";
            listPoints.Add(F6);

            //FC5
            SeriesPoint FC5 = new SeriesPoint(-37.949, 16.869);
            FC5.Tag = "FC5";
            listPoints.Add(FC5);

            //FCZ
            SeriesPoint FCZ = new SeriesPoint(0, 16.869);
            FCZ.Tag = "FCZ";
            listPoints.Add(FCZ);

            //FC6 
            SeriesPoint FC6 = new SeriesPoint(35.391, 16.869);
            FC6.Tag = "FC6";
            listPoints.Add(FC6);

            //T7 
            SeriesPoint T7 = new SeriesPoint(-46, -10);
            T7.Tag = "T7";
            listPoints.Add(T7);

            //C3 
            SeriesPoint C3 = new SeriesPoint(-27, -10);
            C3.Tag = "C3";
            listPoints.Add(C3);

            //Cz
            SeriesPoint Cz = new SeriesPoint(0, -10);
            Cz.Tag = "Cz";
            listPoints.Add(Cz);

            //C4 
            SeriesPoint C4 = new SeriesPoint(27, -10);
            C4.Tag = "C4";
            listPoints.Add(C4);

            //T8
            SeriesPoint T8 = new SeriesPoint(46, -10);
            T8.Tag = "T8";
            listPoints.Add(T8);


            //CP5
            SeriesPoint CP5 = new SeriesPoint(-32, -40);
            CP5.Tag = "CP5";
            listPoints.Add(CP5);

            //CPZ
            SeriesPoint CPZ = new SeriesPoint(0, -40);
            CPZ.Tag = "CPZ";
            listPoints.Add(CPZ);

            //CP6 
            SeriesPoint CP6 = new SeriesPoint(32, -40);
            CP6.Tag = "CP6";
            listPoints.Add(CP6);

            //P5
            SeriesPoint P5 = new SeriesPoint(-25, -62);
            P5.Tag = "P5";
            listPoints.Add(P5);

            //Pz
            SeriesPoint Pz = new SeriesPoint(0, -69);
            Pz.Tag = "Pz";
            listPoints.Add(Pz);
            //P6
            SeriesPoint P6 = new SeriesPoint(25, -62);
            P6.Tag = "P6";
            listPoints.Add(P6);
            //O1
            SeriesPoint O1 = new SeriesPoint(-15.5, -64);
            O1.Tag = "O1";
            listPoints.Add(O1);

            //O2
            SeriesPoint O2 = new SeriesPoint(15.5, -64);
            O2.Tag = "O2";
            listPoints.Add(O2);

            return listPoints.ToArray();
        }


        /// <summary>
        /// Compose 2D geometry data
        /// </summary>
        private void ComposeGeometryAdditionalData2D()
        {
            _gads2D = new GeometryAdditionalData[_dataSizeX][];
            double xStep = (_XMax - _XMin) / (double)(_dataSizeX - 1);
            double yStep = (_YMax + 10 - _YMin - 10) / (double)(_dataSizeY - 1);
            System.Threading.Tasks.Parallel.For(0, _dataSizeX, (i) =>
            {
                double x = _XMin + (double)i * xStep;
                _gads2D[i] = new GeometryAdditionalData[_dataSizeY];

                for (int j = 0; j < _dataSizeY; j++)
                {
                    double y = _YMin - 10 + (double)j * yStep;
                    GeometryAdditionalData gad = new GeometryAdditionalData();
                    gad.DistancesToDataPoint = new double[_EEGchannelsCount];
                    gad.Weights = new double[_EEGchannelsCount];
                    //Use weighted average method 
                    double sumOfWeights = 0;
                    for (int point = 0; point < _EEGchannelsCount; point++)
                    {
                        SeriesPoint p = _sensorLocations[point];
                        double distance = Math.Sqrt((p.X - x) * (p.X - x) + (p.Y - y) * (p.Y - y));
                        gad.DistancesToDataPoint[point] = distance;
                        double weight = 1;
                        if (distance != 0)
                        {
                            //weight = 1.0 / Math.Pow(distance, 3);
                            weight = 1.0 / (distance * distance * distance);
                        }
                        gad.Weights[point] = weight;

                        sumOfWeights += weight;
                    }

                    gad.SumOfWeights = sumOfWeights;

                    _gads2D[i][j] = gad;
                }
            });
        }

        /// <summary>
        /// Buffer for Area series
        /// </summary>
        private void CreateBuffers()
        {
            _buffer = new double[_EEGchannelsCount][];
            _buffer04 = new double[_EEGchannelsCount];
            _buffer48 = new double[_EEGchannelsCount];
            _buffer812 = new double[_EEGchannelsCount];
            _buffer1232 = new double[_EEGchannelsCount];
            _buffer32100 = new double[_EEGchannelsCount];
            for (int i = 0; i < _EEGchannelsCount; i++)
            {
                _buffer[i] = new double[_iWinSizeForFFT];
            }
        }

        /// <summary>
        /// Update head HeatMaps
        /// </summary>
        /// <param name="values"></param>
        /// <param name="Head2DIndex"></param>
        private void UpdateHeadHeatmaps(double[] values, int Head2DIndex)
        {
            int[] colors;
                colors = ChartTools.ConvertDataToColorsByFixedIntervalPalette(values, _paletteStepValues[Head2DIndex], _paletteStepColors);

            // Update chart on the Main UI Thread
            Dispatcher.Invoke(() =>
            {
                if (_running)
                    for (int i = 0; i < colors.Length; i++)
                        _ChartList[3 + Head2DIndex].ViewXY.FreeformPointLineSeries[0].SeriesEventMarkers[i].Symbol.Color1 = FromArgb(colors[i]);
            }, DispatcherPriority.Send);

            CalculateGeometryColors(values, Head2DIndex);
        }

        /// <summary>
        /// Calculate geometry colors
        /// </summary>
        void CalculateGeometryColors(double[] values, int Head2DIndex)
        {
            try
            {
                IntensityPoint[,] data = _ChartList[3 + Head2DIndex].ViewXY.IntensityGridSeries[1].Data;
                System.Threading.Tasks.Parallel.For(0, _dataSizeX, (i) =>
                {

                    GeometryAdditionalData[] gadsForThisCol = _gads2D[i];
                    double[] dataValuesForThisCol = _dataValues[i];
                    uint _A = 128;
                    int[] outputColorsForThisCol = _heatmapColorsData[i];

                    for (int j = 0; j < _dataSizeY; j++)
                    {
                        GeometryAdditionalData gad = gadsForThisCol[j];
                        double[] weights = gad.Weights;
                        double sum = 0;

                        for (int point = 0; point < _sensorLocations.Length; point++)
                        {
                            sum += values[point] * weights[point];
                        }
                        dataValuesForThisCol[j] = sum / gad.SumOfWeights;
                    }
                    int[] colorsForThisColumn;
                        colorsForThisColumn = ChartTools.ConvertDataToColorsByFixedIntervalPalette(dataValuesForThisCol, _paletteStepValues[Head2DIndex], _paletteStepColors);
                    for (int j = 0; j < _dataSizeY; j++)
                    {
                        data[i, j].Color = FromArgb((int)(_A << 24) | (colorsForThisColumn[j] & 0x00FFFFFF));
                    }
                });

                // Update chart on the Main UI Thread
                Dispatcher.Invoke(() =>
                {
                    if (_running)
                        _ChartList[3 + Head2DIndex].ViewXY.IntensityGridSeries[1].InvalidateColorsOnly();
                }, DispatcherPriority.Send);

            }
            catch { }

        }

        /// <summary>
        /// Convert int to Color
        /// </summary>
        /// <param name="color">integer color</param>
        /// <returns>Color</returns>
        internal static Color FromArgb(int color)
        {
            return Color.FromArgb((byte)((color & 0xFF000000) >> 24),
                (byte)((color & 0x00FF0000) >> 16),
                (byte)((color & 0x0000FF00) >> 8),
                (byte)(color & 0x000000FF));
        }

    }
    /// <summary>
    /// Parser for EEG data reading
    /// </summary>
    internal class Reader
    {
        public List<(string, List<int>)> data = new List<(string, List<int>)>();

        public Reader(string path)
        {
            try
            {
                string[] lines = System.IO.File.ReadAllLines(path);
                string[] names = lines[0].Split(',');
                for (int i = 0; i < names.Length; i++)
                {
                    data.Add((names[i], new List<int>()));
                }
                for (int i = 1; i < lines.Length; i++)
                {
                    string[] values = lines[i].Split(',');
                    if (values.Length == 1)
                        break;
                    for (int x = 0; x < names.Length; x++)
                    {
                        data[x].Item2.Add(int.Parse(values[x]));
                    }
                }
            }
            catch (Exception ex)
            {
                ex.ToString();
            }
        }
    }
}
